/*
 * Copyright (c) 2016-2019 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * Authors: Mateusz Moscicki <m.moscicki2@partner.samsung.com>
 *          Karol Lewandowski <k.lewandowsk@samsung.com>
 */

#define LOG_TAG "CRASH_MANAGER"
#include "shared/log.h"
#include "dbus-util.h"
#include "defs.h"

#include <assert.h>
#include <fcntl.h>
#include <getopt.h>
#include <gio/gio.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/procfs.h>
#include <unistd.h>

#define CRASH_BUS_NAME          "org.tizen.system.crash"
#define CRASH_OBJECT_PATH       "/Org/Tizen/System/Crash"
#define CRASH_INTERFACE_NAME    CRASH_BUS_NAME
#define CRASH_PATH_CRASH        CRASH_OBJECT_PATH"/Crash"
#define CRASH_INTERFACE_CRASH   CRASH_INTERFACE_NAME".Crash"
#define PROCESS_CRASHED         "ProcessCrashed"

#define ARM_REG_LR 14
#define ARM_REG_PC 15
#define AARCH64_REG_LR 30

struct RegInfo {
	char *name;
	long long int value;
};

struct NotifyParams {
	bool legacy_notification;
	int prstatus_fd;
	pid_t pid;
	pid_t tid;
	int signal;
	char *cmd_name;
	char *cmd_path;
	char *report_path;
	char *appid;
	char *pkgid;
	char *tid_comm;
};

static int _get_important_registers(int fd, struct RegInfo **reg_info)
{
	int count = -1;

	if (reg_info == NULL) {
		_E("reg_info is NULL\n");
		return -1;
	}

	*reg_info = NULL;
	struct elf_prstatus *prstatus = mmap(NULL, sizeof(struct elf_prstatus),
					 PROT_READ, MAP_SHARED, fd, 0);
	if (prstatus == MAP_FAILED) {
		_E("mmap error: %m\n");
		return -1;
	}

#if (defined __i386__) || (defined __x86_64__)
	count = 1;
#elif (defined __arm__) || (defined __aarch64__)
	count = 2;
#endif

	*reg_info = (struct RegInfo*)malloc(sizeof(struct RegInfo)*count);
	struct RegInfo *reginfo = *reg_info;

	if (*reg_info == NULL) {
		_E("malloc for reg_info error: %m");
		goto out;
	}

#if (defined __i386__) || (defined __x86_64__)
	struct user_regs_struct *g_registers = (struct user_regs_struct*)prstatus->pr_reg;

#if defined(__i386__)
	reginfo[0].name = strdup("x86.eip");
	reginfo[0].value = (long long int)g_registers->eip;
#else
	reginfo[0].name = strdup("x86_64.rip");
	reginfo[0].value = (long long int)g_registers->rip;
#endif // defined(__i386__)

	if (reginfo[0].name == NULL)
		goto strdup_error;

#elif defined(__arm__)

	struct user_regs *g_registers;
	g_registers = (struct user_regs*)prstatus->pr_reg;

	if ((reginfo[0].name = strdup("arm.pc")) == NULL)
		goto strdup_error;
	reginfo[0].value = (long long int)g_registers->uregs[ARM_REG_PC];
	if ((reginfo[1].name = strdup("arm.lr")) == NULL)
		goto strdup_error;
	reginfo[1].value = (long long int)g_registers->uregs[ARM_REG_LR];

#elif defined(__aarch64__)

	struct user_regs_struct *g_registers = (struct user_regs_struct*)prstatus->pr_reg;

	if ((reginfo[0].name = strdup("aarch64.pc")) == NULL)
		goto strdup_error;
	reginfo[0].value = (long long int)g_registers->pc;
	if ((reginfo[1].name = strdup("aarch64.lr")) == NULL)
		goto strdup_error;
	reginfo[1].value = (long long int)g_registers->regs[AARCH64_REG_LR];

#endif // (defined __i386__) || (defined __x86_64__)

out:
	if (count <= 0) {
		free(*reg_info);
		*reg_info = NULL;
	}

	munmap(prstatus, sizeof(*prstatus));
	return count;

strdup_error:
	_E("strdup error: %m\n");
	count = -1;
	goto out;
}

static GVariant* build_legacy_message_data(const struct NotifyParams *notify_params)
{
	assert(notify_params);

	return g_variant_new("(ssss)", notify_params->cmd_name, notify_params->cmd_path, notify_params->appid, notify_params->pkgid);
}

static GVariant* build_message_data(const struct NotifyParams *notify_params)
{
	assert(notify_params);

	GVariantBuilder md_builder;
	g_variant_builder_init(&md_builder, G_VARIANT_TYPE("(sssssiia{sv})"));

	g_variant_builder_add(&md_builder, "s", notify_params->cmd_name);
	g_variant_builder_add(&md_builder, "s", notify_params->cmd_path);
	g_variant_builder_add(&md_builder, "s", notify_params->appid);
	g_variant_builder_add(&md_builder, "s", notify_params->pkgid);
	g_variant_builder_add(&md_builder, "s", notify_params->report_path);
	g_variant_builder_add(&md_builder, "i", notify_params->pid);
	g_variant_builder_add(&md_builder, "i", notify_params->tid);

	g_variant_builder_open(&md_builder, G_VARIANT_TYPE("a{sv}"));

	if (notify_params->signal) {
		g_variant_builder_open(&md_builder, G_VARIANT_TYPE("{sv}"));
		g_variant_builder_add(&md_builder, "s", "sys.signal");
		g_variant_builder_add(&md_builder, "v", g_variant_new_int32(notify_params->signal));
		g_variant_builder_close(&md_builder);
	}

	if (notify_params->tid_comm) {
		g_variant_builder_open(&md_builder, G_VARIANT_TYPE("{sv}"));
		g_variant_builder_add(&md_builder, "s", "sys.tid.comm");
		g_variant_builder_add(&md_builder, "v", g_variant_new_string(notify_params->tid_comm));
		g_variant_builder_close(&md_builder);
	}

	struct RegInfo *reg_info;
	int regs_count = _get_important_registers(notify_params->prstatus_fd, &reg_info);

	/* registers */
	for (int i = 0; i < regs_count; i++) {
		g_variant_builder_open(&md_builder, G_VARIANT_TYPE("{sv}"));

		g_variant_builder_add(&md_builder, "s", reg_info[i].name);
#if (defined __x86_64__) || (defined __aarch64__)
		g_variant_builder_add(&md_builder, "v", g_variant_new_uint64(reg_info[i].value));
#else
		g_variant_builder_add(&md_builder, "v", g_variant_new_uint32(reg_info[i].value));
#endif

		free(reg_info[i].name);
		g_variant_builder_close(&md_builder);
	}
	free(reg_info);

	g_variant_builder_close(&md_builder);

	return g_variant_builder_end(&md_builder);
}

static bool send_notify(GDBusConnection *conn, const struct NotifyParams *notify_params)
{
	int result = false;
	GError *error = NULL;

	GVariant *data = notify_params->legacy_notification
			 ? build_legacy_message_data(notify_params)
			 : build_message_data(notify_params);
	if (data == NULL) {
		_E("Error while preparing parameters");
		goto out;
	}

	g_dbus_connection_emit_signal(conn,
				 NULL,
				 CRASH_PATH_CRASH,
				 CRASH_INTERFACE_CRASH,
				 PROCESS_CRASHED,
				 data,
				 &error);
	if (error) {
		_E("Failed to emit signal: %s", error->message);
		goto out;
	}

	g_dbus_connection_flush_sync(conn, NULL, &error);
	if (error)
		_E("Failed to flush connection - signal might not be delivered: %s", error->message);

	result = true;

out:
	if (error)
		g_error_free(error);

	return result;
}

static bool parse_cmdline(int ac, char *av[], struct NotifyParams *params)
{
	assert(av);
	assert(params);

	enum {
		FLAG_CMDLINE = 1,
		FLAG_CMDPATH,
		FLAG_PID,
		FLAG_TID,
		FLAG_APPID,
		FLAG_PKGID,
		FLAG_REPORTPATH,
		FLAG_PRSTATUS_FD,
		FLAG_SIGNAL,
		FLAG_TID_COMM,
		FLAG_LEGACY_NOTIFICATION,
	};
	static const struct option options[] = {
		{ .name = "cmdline",     .has_arg = required_argument, .flag = NULL, .val = FLAG_CMDLINE },
		{ .name = "cmdpath",     .has_arg = required_argument, .flag = NULL, .val = FLAG_CMDPATH },
		{ .name = "pid",         .has_arg = required_argument, .flag = NULL, .val = FLAG_PID },
		{ .name = "tid",         .has_arg = required_argument, .flag = NULL, .val = FLAG_TID },
		{ .name = "appid",       .has_arg = required_argument, .flag = NULL, .val = FLAG_APPID },
		{ .name = "pkgid",       .has_arg = required_argument, .flag = NULL, .val = FLAG_PKGID },
		{ .name = "reportpath",  .has_arg = required_argument, .flag = NULL, .val = FLAG_REPORTPATH },
		{ .name = "prstatus_fd", .has_arg = required_argument, .flag = NULL, .val = FLAG_PRSTATUS_FD },
		{ .name = "signal",      .has_arg = required_argument, .flag = NULL, .val = FLAG_SIGNAL },
		{ .name = "tid-comm",    .has_arg = required_argument, .flag = NULL, .val = FLAG_TID_COMM },
		{ .name = "legacy-sig",  .has_arg = no_argument,       .flag = NULL, .val = FLAG_LEGACY_NOTIFICATION },
		{ NULL },
	};

	int val;
	do {
		val = getopt_long_only(ac, av, "", options, NULL);

		if (FLAG_CMDLINE == val)
			params->cmd_name = basename(optarg);
		else if (FLAG_CMDPATH == val)
			params->cmd_path = optarg;
		else if (FLAG_REPORTPATH == val)
			params->report_path = optarg;
		else if (FLAG_PID == val)
			params->pid = atoi(optarg);
		else if (FLAG_TID == val)
			params->tid = atoi(optarg);
		else if (FLAG_APPID == val)
			params->appid = optarg;
		else if (FLAG_PKGID == val)
			params->pkgid = optarg;
		else if (FLAG_PRSTATUS_FD == val)
			params->prstatus_fd = atoi(optarg);
		else if (FLAG_SIGNAL == val)
			params->signal = atoi(optarg);
		else if (FLAG_TID_COMM == val)
			params->tid_comm = optarg;
		else if (FLAG_LEGACY_NOTIFICATION == val)
			params->legacy_notification = true;
	} while (val != -1);

	return params->cmd_name && params->cmd_path && params->appid && params->pkgid && params->prstatus_fd > 0;
}

static void usage(const char *const progname)
{
	assert(progname);

	printf("%s --prstatus_fd N --pid PID --tid TID --cmdline CMDLINE --cmdpath CMDPATH --appid APPID --pkgid PKGID --signal SIGNALNR [--legacy-sig]\n", progname);
}

int main(int ac, char *av[])
{
	struct NotifyParams params = {0, };

	if (!parse_cmdline(ac, av, &params)) {
		usage(av[0]);
		return EXIT_FAILURE;
	}

	GDBusConnection *conn = NULL;
	if (!bus_get(&conn))
		return EXIT_FAILURE;

	send_notify(conn, &params);
	bus_put(conn);

	return EXIT_SUCCESS;
}
